/*
 *
 *  OBEX Server
 *
 *  Copyright (C) 2007-2010  Marcel Holtmann <marcel@holtmann.org>
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

struct obex_transport_driver {
	const char *name;
	uint16_t service;
	void *(*start) (struct obex_server *server, int *err);
	int (*getpeername) (GIOChannel *io, char **name);
	int (*getsockname) (GIOChannel *io, char **name);
	void (*stop) (void *data);
};

int obex_transport_driver_register(struct obex_transport_driver *driver);
void obex_transport_driver_unregister(struct obex_transport_driver *driver);
GSList *obex_transport_driver_list(void);
